# Reseda version 1.0.0
# MusetCalculator.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Accessors
# 4) Elements
# 5) Blocks
# 6) Nodes
# 7) Edges
# 8) Sets

# 1) Imported modules
# -------------------

import time

from S2MLCore import *
from ResedaModels import *
from MusetElement import *
from MusetSet import *

# 2) Creator
# ----------

class MusetCalculator:

  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._targetBlock = None
    self._attributePrefix = "@FLAG"
    self._attributeIndex = 0

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) Accessors
# ------------

  def GetTargetBlock(self):
    return self._targetBlock

  def SetTargetBlock(self, block):
    self._targetBlock = block

  def _NewAttributeName(self):
    self._attributeIndex += 1
    return self._attributePrefix + str(self._attributeIndex)

# 4) Elements
# -----------

# 4.1) Parent blocks
# ------------------

  def BuildParentBlocksOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectParentBlocksOfElement(musetElement, musetSet)
    return musetSet

  def CollectParentBlocksOfElement(self, musetElement, musetParentBlocks):
    if musetElement.IsNetworkElement():
      element = musetElement.GetElement()
      parentBlock = element.GetContainer()
      if parentBlock==None:
        return
      if parentBlock.GetContainer()==None:
        return
      if musetParentBlocks.DoesContainBlock(parentBlock):
        return
      musetParentBlock = MusetElement.NewBlock(parentBlock)
      musetParentBlocks.AppendElement(musetParentBlock)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectParentBlocksOfElement(musetChildElement, musetParentBlocks)

# 4.2) Ancestors blocks
# ---------------------

  def BuildAncestorBlocksOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectAncestorBlocksOfElement(musetElement, musetSet)
    return musetSet

  def CollectAncestorBlocksOfElement(self, musetElement, musetAncestorBlocks):
    if musetElement.IsNetworkElement():
      element = musetElement.GetNetworkElement()
      parentBlock = element.GetContainer()
      while True:
        if parentBlock==None:
          break
        if parentBlock.GetContainer()==None:
          break
        if musetAncestorBlocks.DoesContainBlock(parentBlock):
          break
        musetParentBlock = MusetElement.NewBlock(parentBlock)
        musetAncestorBlocks.AppendElement(musetParentBlock)
        parentBlock = parentBlock.GetContainer()
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectAncestorBlocksOfElement(musetChildElement, musetAncestorBlocks)

# 4.3) Child blocks
# -----------------

  def BuildChildBlocksOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectChildBlocksOfElement(musetElement, musetSet)
    return musetSet

  def CollectChildBlocksOfElement(self, musetElement, musetChildBlocks):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      for element in block.GetElements():
        if ResedaType.IsBlock(element) and not musetChildBlocks.DoesContainBlock(element):
          musetChildBlock = MusetElement.NewBlock(element)
          musetChildBlocks.AppendElement(musetChildBlock)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectChildBlocksOfElement(musetChildElement, musetChildBlocks)

# 4.4) Descendant blocks
# ----------------------

  def BuildDescendantBlocksOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectDescendantBlocksOfElement(musetElement, musetSet)
    return musetSet

  def CollectDescendantBlocksOfElement(self, musetElement, musetDescendantBlocks):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      self.CollectDescendantBlocksOfBlock(block, musetDescendantBlocks)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectDescendantBlocksOfElement(musetChildElement, musetDescendantBlocks)

  def CollectDescendantBlocksOfBlock(self, block, musetDescendantBlocks):
    for element in block.GetElements():
      if ResedaType.IsBlock(element):
        if not musetDescendantBlocks.DoesContainBlock(element):
          musetDescendantBlock = MusetElement.NewBlock(element)
          musetDescendantBlocks.AppendElement(musetDescendantBlock)
        self.CollectDescendantBlocksOfBlock(element, musetDescendantBlocks)

# 4.5) Child nodes
# ----------------

  def BuildChildNodesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectChildNodesOfElement(musetElement, musetSet)
    return musetSet

  def CollectChildNodesOfElement(self, musetElement, musetChildNodes):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      for element in block.GetElements():
        if ResedaType.IsNode(element) and not musetChildNodes.DoesContainNode(element):
          musetChildNode = MusetElement.NewNode(element)
          musetChildNodes.AppendElement(musetChildNode)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectChildNodesOfElement(musetChildElement, musetChildNodes)

# 4.6) Descendant nodes
# ---------------------

  def BuildDescendantNodesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectDescendantNodesOfElement(musetElement, musetSet)
    return musetSet

  def CollectDescendantNodesOfElement(self, musetElement, musetDescendantNodes):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      self.CollectDescendantNodesOfBlock(block, musetDescendantNodes)
      return
    if musetElement.IsNode():
      node = musetElement.GetNode()
      if not musetDescendantNodes.DoesContainNode(node):
        musetDescendantNode = MusetElement.NewNode(node)
        musetDescendantNodes.AppendElement(musetDescendantNode)
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectDescendantNodesOfElement(musetChildElement, musetDescendantNodes)

  def CollectDescendantNodesOfBlock(self, block, musetDescendantNodes):
    for element in block.GetElements():
      if ResedaType.IsBlock(element):
       self.CollectDescendantNodesOfBlock(element, musetDescendantNodes)
      elif ResedaType.IsNode(element):
        if not musetDescendantNodes.DoesContainNode(element):
          musetDescendantNode = MusetElement.NewNode(element)
          musetDescendantNodes.AppendElement(musetDescendantNode)
 
# 4.7) Child edges
# ----------------

  def BuildChildEdgesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectChildEdgesOfElement(musetElement, musetSet)
    return musetSet

  def CollectChildEdgesOfElement(self, musetElement, musetChildEdges):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      for element in block.GetElements():
        if ResedaType.IsEdge(element) and not musetChildNodes.DoesContainEdge(element):
          musetChildEdge = MusetElement.NewEdge(element)
          musetChildEdges.AppendElement(musetChildEdge)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectChildEdgesOfElement(musetChildElement, musetChildEdges)

# 4.8) Descendant edges
# ---------------------

  def BuildDescendantEdgesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectDescendantEdgesOfElement(musetElement, musetSet)
    return musetSet

  def CollectDescendantEdgesOfElement(self, musetElement, musetDescendantEdges):
    if musetElement.IsBlock():
      block = musetElement.GetBlock()
      self.CollectDescendantEdgesOfBlock(block, musetDescendantEdges)
      return
    if musetElement.IsEdge():
      edge = musetElement.GetEdge()
      if not musetDescendantEdges.DoesContainEdge(edge):
        musetDescendantEdge = MusetElement.NewEdge(edge)
        musetDescendantEdges.AppendElement(musetDescendantEdge)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectDescendantEdgesOfElement(musetChildElement, musetDescendantEdges)

  def CollectDescendantEdgesOfBlock(self, block, musetDescendantEdges):
    for element in block.GetElements():
      if ResedaType.IsBlock(element):
       self.CollectDescendantEdgesOfBlock(element, musetDescendantEdges)
      elif ResedaType.IsEdge(element):
        if not musetDescendantEdges.DoesContainEdge(element):
          musetDescendantEdge = MusetElement.NewEdge(element)
          musetDescendantEdges.AppendElement(musetDescendantEdge)

# 5) Blocks
# ---------


# 6) Nodes
# --------

  def BuildInEdgesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectInEdgesOfElement(musetElement, musetSet)
    return musetSet

  def CollectInEdgesOfElement(self, musetElement, musetInEdges):
    if musetElement.IsNode():
      node = musetElement.GetNode()
      for edge in node.GetInEdges():
        if not musetInEdges.DoesContainEdge(edge):
          musetInEdge = MusetElement.NewEdge(edge)
          musetInEdges.AppendElement(musetInEdge)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectInEdgesOfElement(musetChildElement, musetInEdges)

  def BuildOutEdgesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectOutEdgesOfElement(musetElement, musetSet)
    return musetSet

  def CollectOutEdgesOfElement(self, musetElement, musetOutEdges):
    if musetElement.IsNode():
      node = musetElement.GetNode()
      for edge in node.GetOutEdges():
        if not musetOutEdges.DoesContainEdge(edge):
          musetOutEdge = MusetElement.NewEdge(edge)
          musetOutEdges.AppendElement(musetOutEdge)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectOutEdgesOfElement(musetChildElement, musetOutEdges)

  def BuildReachableNodesOfElement(self, musetElement, testOnAttributesOfNodes, testOnAttributesOfEdges):
    musetSet = MusetSet.NewEmptySet()
    self.CollectReachableNodesOfElement(musetElement, musetSet, testOnAttributesOfNodes, testOnAttributesOfEdges)
    return musetSet

  def CollectReachableNodesOfElement(self, musetElement, musetReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges):
    flagName = self._NewAttributeName()
    self._CollectReachableNodesOfElement(musetElement, musetReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName)
    for musetNode in musetReachableNodes.GetElements():
      node = musetNode.GetNode()
      node.ResetAttribute(flagName)

  def _CollectReachableNodesOfElement(self, musetElement, musetReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName):
    if musetElement.IsNode():
      sourceNode = musetElement.GetNode()
      if sourceNode.GetAttribute(flagName):
        return
      satisfied = self._engine.CalculateValueOfTestOnAttributes(sourceNode, testOnAttributesOfNodes)
      if not satisfied:
        return
      candidates = [sourceNode]
      while len(candidates)>0:
        sourceNode = candidates.pop()
        sourceNode.SetAttribute(flagName, True)
        musetReachableNode = MusetElement.NewNode(sourceNode)
        musetReachableNodes.AppendElement(musetReachableNode)
        for edge in sourceNode.GetOutEdges():
          satisfied = self._engine.CalculateValueOfTestOnAttributes(edge, testOnAttributesOfEdges)
          if not satisfied:
            continue
          targetNode = edge.GetTargetNode()
          if targetNode==sourceNode:
            targetNode = edge.GetSourceNode()
          satisfied = self._engine.CalculateValueOfTestOnAttributes(targetNode, testOnAttributesOfNodes)
          if not satisfied:
            continue
          if not targetNode.GetAttribute(flagName):
            targetNode.SetAttribute(flagName, True)
            candidates.append(targetNode)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self._CollectReachableNodesOfElement(musetChildElement, musetReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName)

  def BuildCoReachableNodesOfElement(self, musetElement, testOnAttributesOfNodes, testOnAttributesOfEdges):
    musetSet = MusetSet.NewEmptySet()
    self.CollectCoReachableNodesOfElement(musetElement, musetSet, testOnAttributesOfNodes, testOnAttributesOfEdges)
    return musetSet

  def CollectCoReachableNodesOfElement(self, musetElement, musetCoReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges):
    flagName = self._NewAttributeName()
    self._CollectCoReachableNodesOfElement(musetElement, musetCoReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName)
    for musetNode in musetCoReachableNodes.GetElements():
      node = musetNode.GetNode()
      node.ResetAttribute(flagName)

  def _CollectCoReachableNodesOfElement(self, musetElement, musetCoReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName):
    if musetElement.IsNode():
      targetNode = musetElement.GetNode()
      if targetNode.GetAttribute(flagName):
        return
      satisfied = self._engine.CalculateValueOfTestOnAttributes(targetNode, testOnAttributesOfNodes)
      if not satisfied:
        return
      candidates = [targetNode]
      while len(candidates)>0:
        targetNode = candidates.pop()
        targetNode.SetAttribute(flagName, True)
        musetCoReachableNode = MusetElement.NewNode(targetNode)
        musetCoReachableNodes.AppendElement(musetCoReachableNode)
        for edge in targetNode.GetInEdges():
          satisfied = self._engine.CalculateValueOfTestOnAttributes(edge, testOnAttributesOfEdges)
          if not satisfied:
            continue
          sourceNode = edge.GetSourceNode()
          if sourceNode==targetNode:
            sourceNode = edge.GetTargetNode()
          satisfied = self._engine.CalculateValueOfTestOnAttributes(sourceNode, testOnAttributesOfNodes)
          if not satisfied:
            continue
          if not sourceNode.GetAttribute(flagName):
            sourceNode.SetAttribute(flagName, True)
            candidates.append(sourceNode)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self._CollectCoReachableNodesOfElement(musetChildElement, musetCoReachableNodes, testOnAttributesOfNodes, testOnAttributesOfEdges, flagName)

# 7) Edges
# --------

  def BuildSourceNodesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectSourceNodesOfElement(musetElement, musetSet)
    return musetSet

  def CollectSourceNodesOfElement(self, musetElement, musetSourceNodes):
    if ResedaType.IsEdge(musetElement) or ResedaType.IsArc(musetElement):
      edge = musetEdge.GetEdge()
      sourceNode = edge.GetSourceNode()
      if not musetSourceNodes.DoesContainNode(sourceNode):
        musetSourceNode = MusetElement.NewNode(sourceNode)
        musetSourceNodes.AppendElement(musetSourceNode)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectSourceNodesOfElement(musetChildElement, musetSourceNodes)

  def BuildTargetNodesOfElement(self, musetElement):
    musetSet = MusetSet.NewEmptySet()
    self.CollectTargetNodesOfElement(musetElement, musetSet)
    return musetSet

  def CollectTargetNodesOfElement(self, musetElement, musetTargetNodes):
    if ResedaType.IsEdge(musetElement) or ResedaType.IsArc(musetElement):
      edge = musetEdge.GetEdge()
      targetNode = edge.GetTargetNode()
      if not musetTargetNodes.DoesContainNode(targetNode):
        musetTargetNode = MusetElement.NewNode(targetNode)
        musetTargetNodes.AppendElement(musetTargetNode)
      return
    if musetElement.IsSet():
      for musetChildElement in musetElement.GetElements():
        self.CollectTargetNodesOfElement(musetChildElement, musetTargetNodes)

# 8) Sets
# -------

  def BuildUnion(self, set1, set2):
    result = MusetSet()
    for element in set1._elements:
      result._elements.append(element)
    for element in set2._elements:
      if not result.DoesContainElement(element):
        result._elements.append(element)
    return result

  def BuildIntersection(self, set1, set2):
    result = MusetSet()
    for element in set1._elements:
      if set2.DoesContainElement(element):
        result._elements.append(element)
    return result

  def BuildDifference(self, set1, set2):
    result = MusetSet()
    for element in set1._elements:
      if not set2.DoesContainElement(element):
        result._elements.append(element)
    return result

  def BuildSelection(self, set, test):
    result = MusetSet()
    for element in set._elements:
      if test(element):
        result._elements.append(element)
    return result

# 9) Attributes
# -------------


  def SetAttributeOfElements(self, musetSet, musetName, musetValue):
    if not musetSet.IsSet() or not musetName.IsSymbolicConstant() or not musetValue.IsConstant():
      return
    constant = musetName.GetConstant()
    name = constant.GetValue()
    value = musetValue.GetConstant()
    for musetElement in musetSet.GetMembers():
      if not musetElement.IsElement():
        continue
      element = musetElement.GetElement()
      expression = S2MLStochasticExpression.NewConstant(value)
      element.SetAttribute(name, expression)

  def ResetAttributeOfElements(self, musetSet, musetName):
    if not musetSet.IsSet() or not musetName.IsSymbolicConstant():
      return
    constant = musetName.GetConstant()
    name = constant.GetValue()
    value = musetValue.GetConstant()
    for musetElement in musetSet.GetMembers():
      if not musetElement.IsElement():
        continue
      element = musetElement.GetElement()
      element.ResetAttribute(name)

