# Reseda version 1.0.0
# ResedaShortestPathCalculator.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Disjkra algorithm

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLCore import *
from ResedaModels import *

# 2) Creator
# ----------

class ResedaShortestPathCalculator:
  LENGTH = "@LENGTH"
  DISTANCE = "@DISTANCE"
  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._expressionManager = self._engine.GetExpressionManager()
    self._targetBlock = None
    self._sourceNode = None
    self._targetNode = None
    self._matrix = None

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) Disjkra algorithm
# --------------------

  def ComputeShortestPathWithDijskraAlgorithm(self, block, sourceNodeName, targetNodeName, lengthName):
    self._targetBlock = block
    if self._targetBlock==None:
      self.RaiseError("No target block")
      return None
    sourceNodePath = S2MLPath.FromString(sourceNodeName)
    self._sourceNode = self._targetBlock.AccessNode(sourceNodePath)
    if self._sourceNode==None:
      self.RaiseError("No node %s" % sourceNodeName)
      return None
    targetNodePath = S2MLPath.FromString(targetNodeName)
    self._targetNode = self._targetBlock.AccessNode(targetNodePath)
    if self._targetNode==None:
      self.RaiseError("No node %s" % targetNodeName)
      return None
    error = self.InitializeGraphToComputeShortestPath(lengthName)
    if error:
      return None
    self.CalculateDistances()
    shortestPath = []
    found = self.GetShortestPath(self._sourceNode, shortestPath)
    if not found:
      shortestPath = None
    matrix = self.CreateResultMatrixForShortestPath(shortestPath)
    self._targetBlock.ResetAttributesOfElements(ResedaModel.IsEdge, ResedaShortestPathCalculator.LENGTH, False)
    self._targetBlock.ResetAttributesOfElements(ResedaModel.IsNode, ResedaShortestPathCalculator.DISTANCE, False)
    return matrix

  def InitializeGraphToComputeShortestPath(self, lengthName):
    self._targetBlock.BuildGraph()
    self._expressionManager.ResetValuesOfParameters(self._targetBlock)
    for edge in self._targetBlock.GetEdges():
      expression = edge.GetAttribute(lengthName)
      if expression==None:
        self.RaiseError("Attribute %s missing in edge %s." % (lengthName, edge.GetIdentifier()))
        return 1
      else:
        value = self._expressionManager.CalculateValueOfExpression(expression)
        if not value.IsNumber():
          self.RaiseError("Attribute %s of edge %s is not a number" % (lengthName, edge.GetIdentifier()))
          return 1
        length = value.ToFloat()
        if length<0:
          self.RaiseError("Attribute %s of edge %s has a negative value" % (lengthName, edge.GetIdentifier()))
          return 1
        edge.SetAttribute(ResedaShortestPathCalculator.LENGTH, length)
    for node in self._targetBlock.GetNodes():
      node.SetAttribute(ResedaShortestPathCalculator.DISTANCE, math.inf)
    return 0

  def CalculateDistances(self):
    self._sourceNode.SetAttribute(ResedaShortestPathCalculator.DISTANCE, 0.0)
    candidates = [self._sourceNode]
    while candidates:
      node = self.SelectCandidate(candidates)
      if node==self._targetNode:
        break
      candidates.remove(node)
      for edge in node.GetOutEdges():
        nodeDistance = node.GetAttribute(ResedaShortestPathCalculator.DISTANCE)
        oppositeNode = edge.GetOppositeNode(node)
        oppositeNodeDistance = oppositeNode.GetAttribute(ResedaShortestPathCalculator.DISTANCE)
        edgeLength = edge.GetAttribute(ResedaShortestPathCalculator.LENGTH)
        if oppositeNodeDistance==math.inf:
          oppositeNode.SetAttribute(ResedaShortestPathCalculator.DISTANCE, nodeDistance + edgeLength)
          candidates.append(oppositeNode)
        elif nodeDistance + edgeLength < oppositeNodeDistance:
          oppositeNode.SetAttribute(ResedaShortestPathCalculator.DISTANCE, nodeDistance + edgeLength)

  def SelectCandidate(self, candidates):
    shortestDistance = None
    selectedNode = None
    for node in candidates:
      nodeDistance = node.GetAttribute(ResedaShortestPathCalculator.DISTANCE)
      if shortestDistance==None:
        selectedNode = node
        shortestDistance = nodeDistance
      elif nodeDistance < shortestDistance:
        selectedNode = node
        shortestDistance = nodeDistance
    return selectedNode

  def GetShortestPath(self, node, path):
    if node==self._targetNode:
      return True
    nodeDistance = node.GetAttribute(ResedaShortestPathCalculator.DISTANCE)
    for edge in node.GetOutEdges():
      oppositeNode = edge.GetOppositeNode(node)
      oppositeNodeDistance = oppositeNode.GetAttribute(ResedaShortestPathCalculator.DISTANCE)
      edgeLength = edge.GetAttribute(ResedaShortestPathCalculator.LENGTH)
      if oppositeNodeDistance==nodeDistance + edgeLength:
        path.append(edge)
        found = self.GetShortestPath(oppositeNode, path)
        if found:
          return True
        path.pop()
    return False

  def CreateResultMatrixForShortestPath(self, shortestPath):
    matrix = S2MLMatrix()
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Model"))
    row.Put(1, S2MLConstant.NewString(self._targetBlock.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("source"))
    row.Put(1, S2MLConstant.NewString(self._sourceNode.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("target"))
    row.Put(1, S2MLConstant.NewString(self._targetNode.GetIdentifier()))
    if shortestPath==None:
      row = matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("No path"))
      return
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("edge"))
    row.Put(1, S2MLConstant.NewString("target"))
    row.Put(2, S2MLConstant.NewString("distance"))
    targetNode = self._sourceNode
    for edge in shortestPath:
      row = matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString(edge.GetIdentifier()))
      if edge.GetSourceNode()==targetNode:
        targetNode = edge.GetTargetNode()
      else:
        targetNode = edge.GetSourceNode()
      row.Put(1, S2MLConstant.NewString(targetNode.GetIdentifier()))
      row.Put(2, S2MLConstant.NewReal(targetNode.GetAttribute(ResedaShortestPathCalculator.DISTANCE)))
    return matrix


