# S2MLCore version 1.2.0
# S2MLModelParser.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Models
#   3.2) Domains
#   3.3) Blocks and classes
#   3.4) Directives
#   3.5) Attributes
#   3.6) Parameters and stochastic expressions
#   3.7) Functions and operators
# 4) Auxiliary functions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLModels import *
from S2MLModelLexer import *

# 2) Creator
# ----------

class S2MLModelParser:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._lexer = self._engine.GetModelManager().GetModelLexer()
    self._currentTokenIndex = 0
    self._model = None
    self._numberOfErrors = 0

# 3) Models
# ---------

# 3.1) Models
# -----------

  def ImportFile(self, model, fileName):
    try:
      file = open(fileName, "r")
    except:
      self._engine.RaiseError('Unable to open file "%s"' % fileName)
      return 1
    self.ParseFile(model, fileName, file)
    file.close()
    return self._numberOfErrors

  def ParseFile(self, model, fileName, file):
    self._numberOfErrors = 0
    self._numberOfErrors = self._lexer.Initialize(fileName, file)
    if self._numberOfErrors!=0:
      return
    self._ParseModel(model)
    self._lexer.Finalize()
    return self._numberOfErrors

  def _ParseModel(self, model):
    self._model = model
    self._currentTokenIndex = 0
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      elif token.GetType()==S2MLToken.KEYWORD_DOMAIN:
        self._ParseDomainDeclaration(self._model)
      elif token.GetType()==S2MLToken.KEYWORD_BLOCK:
        self._ParseBlock(self._model)
      elif token.GetType()==S2MLToken.KEYWORD_CLASS:
        self._ParseClass(self._model)
      elif token.GetType()==S2MLToken.KEYWORD_FUNCTION:
        self._ParseFunction(self._model)
      elif token.GetType()==S2MLToken.KEYWORD_OPERATOR:
        self._ParseOperator(self._model)
      else:
        self._RaiseParseError(token, "A declaration of class or block was expected ")
        self._currentTokenIndex += 1

# 3.2) Domains
# ------------

  # This method can possibly be derived
  def _ParseDomainDeclaration(self, container):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_DOMAIN)
    if not self._model.CanComposeDomain(container):
      self._RaiseParseError(token, "Domains are not authorized here")
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    domain = self._ParseActualDomain()
    if domain==None:
      return None
    declaration = self._model.NewDomainDeclaration(identifier, domain, container)

  def _ParseActualDomain(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_LBRACE:
      return self._ParseEnumeration()
    if token.GetType()==S2MLToken.TOKEN_LBRACK:
      return self._ParseRange()
    self._RaiseParseError(token, "An enumeration or a range was expected")
    return None

  def _ParseEnumeration(self):
    domain = self._model.NewEnumerationDomain()
    token = self._ParseGivenToken(S2MLToken.TOKEN_LBRACE)
    if token==None:
      return None
    while True:
      constant = self._ParseSymbolicConstant()
      if constant==None:
        return None
      domain.AppendConstant(constant)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()!=S2MLToken.TOKEN_COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(S2MLToken.TOKEN_RBRACE)
    if token==None:
      return None
    return domain

  def _ParseSymbolicConstant(self):
    symbol = self._ParseIdentifier()
    if symbol==None:
      return None
    constant = self._model.NewSymbol(symbol)
    return constant

  def _ParseConstant(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return None
    if token.GetType()==S2MLToken.TOKEN_INTEGER:
      self._currentTokenIndex += 1
      return self._model.NewInteger(int(token.GetString()))
    if token.GetType()==S2MLToken.TOKEN_REAL:
      self._currentTokenIndex += 1
      return self._model.NewReal(float(token.GetString()))
    if token.GetType()==S2MLToken.KEYWORD_TRUE:
      self._currentTokenIndex += 1
      return self._model.NewBoolean(True)
    if token.GetType()==S2MLToken.KEYWORD_FALSE:
      self._currentTokenIndex += 1
      return self._model.NewBoolean(False)
    if token.GetType()==S2MLToken.TOKEN_IDENTIFIER:
      self._currentTokenIndex += 1
      return self._model.NewSymbol(token.GetString())
    self._RaiseParseError(token, "A constant was expected")
    return None

  def _ParseRange(self):
    firstToken = self._ParseGivenToken(S2MLToken.TOKEN_LBRACK)
    token = self._ParseGivenToken(S2MLToken.TOKEN_INTEGER)
    if token==None:
      return None
    lowerBound = int(token.GetString())
    token = self._ParseGivenToken(S2MLToken.TOKEN_COMMA)
    if token==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_INTEGER)
    if token==None:
      return None
    upperBound = int(token.GetString())
    token = self._ParseGivenToken(S2MLToken.TOKEN_RBRACK)
    if token==None:
      return None
    if lowerBound>upperBound:
      self._RaiseParseError(firstToken, "Lower bound greater than upperBound")
      return None
    return self._model.NewRangeDomain(lowerBound, upperBound)

  def _ParseDomain(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.KEYWORD_BOOLEAN:
      self._currentTokenIndex += 1
      return self._model.NewBooleanDomain()
    if token.GetType()==S2MLToken.KEYWORD_INTEGER:
      self._currentTokenIndex += 1
      return self._model.NewIntegerDomain()
    if token.GetType()==S2MLToken.KEYWORD_REAL:
      self._currentTokenIndex += 1
      return self._model.NewRealDomain()
    if token.GetType()==S2MLToken.KEYWORD_SYMBOL:
      self._currentTokenIndex += 1
      return self._model.NewSymbolDomain()
    if token.GetType()==S2MLToken.TOKEN_IDENTIFIER:
      self._currentTokenIndex += 1
      identifier = token.GetString()
      declaration = self._model.LookForIdentifiedElement(identifier)
      if declaration==None:
        self._RaiseParseError(token, "Undeclared domain %s" % identifier)
        return None
      if declaration.GetType()!=S2MLType.DOMAIN_DECLARATION:
        self._RaiseParseError(token, "Not a domain %s" % identifier)
        return None
      return self._model.NewReferenceDomain(identifier)
    return self._ParseActualDomain()

# 3.3) Blocks & classes
# ---------------------

  # This method can possibly be derived
  def _ParseBlock(self, container):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_BLOCK)
    if not self._model.CanComposeBlock(container):
      self._RaiseParseError(token, "Blocks are not authorized here")
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    path = self._ParsePath()
    if path==None:
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    blockDeclaration = self._model.NewBlock(path, container)
    return self._ParseBlockBody(blockDeclaration)

  # This method can possibly be derived
  def _ParseClass(self, container):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_CLASS)
    if not self._model.CanComposeClass(container):
      self._RaiseParseError(token, "Classes are not authorized here")
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    classDeclaration = self._model.NewClass(identifier, container)
    return self._ParseBlockBody(classDeclaration)

  def _ParseBlockBody(self, container):
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()==S2MLToken.KEYWORD_END:
        break
      blockElement = self._ParseBlockElement(container)
      if blockElement==None:
        self._RecoverError(S2MLToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_END)
    if token==None:
      return None
    return container

  # This method must be derived
  def _ParseBlockElement(self, container):
    return None

# 3.4) Directives
# ---------------

  def _ParseClassInstance(self, container):
    if not self._model.CanComposeInstance(container):
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "Instances are not authorized here")
      return None
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      return None
    instantiatedClass = self._model.LookForIdentifiedElement(classIdentifier)
    if instantiatedClass==None:
      self._RaiseParseError(token, "Undeclared class %s" % classIdentifier)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewInstance(identifier, classIdentifier, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_SEMICOLON:
      token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

  def _ParseClonesDirective(self, container):
    if not self._model.CanComposeClonesDirective(container):
      self._RaiseParseError(token, "Clones directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_CLONES)
    pathToClonedElement = self._ParsePath()
    if pathToClonedElement==None:
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_AS)
    if token==None:
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewClonesDirective(identifier, pathToClonedElement, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_SEMICOLON:
      token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

  def _ParseExtendsDirective(self, container):
    if not self._model.CanComposeExtendsDirective(container):
      self._RaiseParseError(token, "Extends directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_EXTENDS)
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    directive = self._model.NewExtendsDirective(classIdentifier, container)
    return directive

  def _ParseEmbedsDirective(self, container):
    if not self._model.CanComposeEmbedsDirective(container):
      self._RaiseParseError(token, "Embeds directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_EMBEDS)
    pathToEmbeddedElement = self._ParsePath()
    if pathToEmbeddedElement==None:
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_AS)
    if token==None:
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewEmbedsDirective(identifier, pathToEmbeddedElement, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_SEMICOLON:
      token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

# 3.5) Attributes
# ---------------

  def _ParseAttributeDirective(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if not self._model.CanComposeAttributeDirective(container):
      self._RaiseParseError(token, "Attribute directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_ATTRIBUTE)
    if token==None:
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    directive = self._model.NewAttributeDirective(path, expression, container)
    return directive

  def _ParseAttributeList(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_LPAR:
      return []
    self._currentTokenIndex += 1
    attributes = []
    while True:
      identifier = self._ParseIdentifier()
      if identifier==None:
        return None
      token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
      if token==None:
        return None
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      attributes.append((identifier, expression))
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLToken.TOKEN_RPAR:
        break
      token = self._ParseGivenToken(S2MLToken.TOKEN_COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    return attributes

  def _ParseAttributes(self, elements):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_LPAR:
      return elements
    self._currentTokenIndex += 1
    while True:
      identifier = self._ParseIdentifier()
      if identifier==None:
        return None
      token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
      if token==None:
        return None
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      for element in elements:
        element.SetAttribute(identifier, expression)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLToken.TOKEN_RPAR:
        break
      token = self._ParseGivenToken(S2MLToken.TOKEN_COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    return elements

# 3.6) Parameters and stochastic expressions
# ------------------------------------------

  def _ParseParameter(self, container):
    if not self._model.CanComposeParameter(container):
      self._RaiseParseError(token, "Parameter declarations are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_PARAMETER)
    if token==None:
      return None
    domain = self._ParseDomain()
    if domain==None:
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    parameter = self._model.NewParameter(domain, path, container)
    parameter.SetDefinition(expression)
    return parameter

  def _ParseStochasticExpression(self):
    expression = self._ParseStochasticExpression8()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.KEYWORD_OR:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.KEYWORD_OR:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression8()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.OR_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression8(self):
    expression = self._ParseStochasticExpression7()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.KEYWORD_AND:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.KEYWORD_AND:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression7()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.AND_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression7(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.KEYWORD_NOT:
      expression = self._ParseStochasticExpression6()
      return expression
    self._currentTokenIndex += 1
    child = self._ParseStochasticExpression6()
    if child==None:
      return None
    expression = S2MLStochasticExpression.NewOperation(S2MLType.NOT_EXPRESSION, [child])
    return expression

  def _ParseStochasticExpression6(self):
    leftMember = self._ParseStochasticExpression5()
    if leftMember==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_EQ:
      self._currentTokenIndex += 1
      type = S2MLType.EQ_EXPRESSION
    elif token.GetType()==S2MLToken.TOKEN_DF:
      self._currentTokenIndex += 1
      type = S2MLType.DF_EXPRESSION
    elif token.GetType()==S2MLToken.TOKEN_LT:
      self._currentTokenIndex += 1
      type = S2MLType.LT_EXPRESSION
    elif token.GetType()==S2MLToken.TOKEN_GT:
      self._currentTokenIndex += 1
      type = S2MLType.GT_EXPRESSION
    elif token.GetType()==S2MLToken.TOKEN_LEQ:
      self._currentTokenIndex += 1
      type = S2MLType.LEQ_EXPRESSION
    elif token.GetType()==S2MLToken.TOKEN_GEQ:
      self._currentTokenIndex += 1
      type = S2MLType.GEQ_EXPRESSION
    else:
      return leftMember
    rightMember = self._ParseStochasticExpression5()
    if rightMember==None:
      return None
    expression = S2MLStochasticExpression.NewOperation(type, [leftMember, rightMember])
    return expression

  def _ParseStochasticExpression5(self):
    expression = self._ParseStochasticExpression4()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_PLUS:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_PLUS:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression4()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.ADD_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression4(self):
    expression = self._ParseStochasticExpression3()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_MINUS:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_MINUS:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression3()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.SUB_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression3(self):
    expression = self._ParseStochasticExpression2()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_STAR:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_STAR:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression2()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.MUL_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression2(self):
    expression = self._ParseStochasticExpression1()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_SLASH:
      return expression
    children = []
    children.append(expression)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_SLASH:
        break
      self._currentTokenIndex += 1
      child = self._ParseStochasticExpression1()
      if child==None:
        return None
      children.append(child)
    expression = S2MLStochasticExpression.NewOperation(S2MLType.DIV_EXPRESSION, children)
    return expression

  def _ParseStochasticExpression1(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLToken.TOKEN_MINUS:
      expression = self._ParseStochasticExpression0()
      return expression
    self._currentTokenIndex += 1
    child = self._ParseStochasticExpression0()
    if child==None:
      return None
    expression = S2MLStochasticExpression.NewOperation(S2MLType.NEG_EXPRESSION, [child])
    return expression

  def _ParseStochasticExpression0(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_LPAR:
      token = self._ParseGivenToken(S2MLToken.TOKEN_LPAR)
      if token==None:
        return None
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
      if token==None:
        return None
      return expression
    elif token.GetType()==S2MLToken.TOKEN_IDENTIFIER:
      identifier = self._ParseIdentifier()
      symbol = self._model.GetSymbol(identifier)
      if symbol!=None:
        expression = S2MLStochasticExpression.NewConstant(S2MLConstant.Copy(symbol))
        return expression
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_LPAR:
        path = S2MLPath.NewEmptyPath()
        path.AppendIdentifier(identifier)
        expression = S2MLStochasticExpression.NewReference(path)
        return expression
      children = self._ParseStochasticExpressionList()
      if children==None:
        return None
      expression = S2MLStochasticExpression.NewOperatorCall(identifier, children)
      return expression
    if token.GetType()==S2MLToken.TOKEN_PATH:
      path = self._ParsePath()
      expression = S2MLStochasticExpression.NewReference(path)
      return expression
    if token.GetType()==S2MLToken.TOKEN_INTEGER:
      token = self._ParseGivenToken(S2MLToken.TOKEN_INTEGER)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewInteger(int(token.GetString())))
      return expression
    if token.GetType()==S2MLToken.TOKEN_REAL:
      token = self._ParseGivenToken(S2MLToken.TOKEN_REAL)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewReal(float(token.GetString())))
      return expression
    if token.GetType()==S2MLToken.KEYWORD_FALSE:
      token = self._ParseGivenToken(S2MLToken.KEYWORD_FALSE)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewBoolean(False))
      return expression
    if token.GetType()==S2MLToken.KEYWORD_TRUE:
      token = self._ParseGivenToken(S2MLToken.KEYWORD_TRUE)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewBoolean(True))
      return expression
    if token.GetType()==S2MLToken.KEYWORD_PI:
      token = self._ParseGivenToken(S2MLToken.KEYWORD_PI)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewReal(math.pi))
      return expression
    if token.GetType()==S2MLToken.KEYWORD_IF:
      return self._ParseIfThenElseExpression()
    if token.GetType()==S2MLToken.KEYWORD_WITHIN:
      return self._ParseWithinExpression()
    operator = self._engine.GetExpressionManager().GetOperatorOfKeywordCode(token.GetType())
    if operator==None:
      self._RaiseParseError(token, "Unknown operator")
      return None
    self._currentTokenIndex += 1
    if operator.IsEmpiricalDistribution():
      distribution = self._ParseDistribution()
      if distribution==None:
        return None
    children = self._ParseStochasticExpressionList()
    if children==None:
      return None
    if len(children)<operator.GetMinimumNumberOfArguments():
      self._RaiseParseError(token, "Too few arguments")
      return None
    if operator.GetMaximumNumberOfArguments()!=S2MLStochasticOperator.MANY and len(children)>operator.GetMaximumNumberOfArguments():
      self._RaiseParseError(token, "Too many arguments")
      return None
    expression = S2MLStochasticExpression.NewOperation(operator.GetType(), children)
    if operator.IsEmpiricalDistribution():
      expression.SetDistribution(distribution)
    return expression

  def _ParseIfThenElseExpression(self):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_IF)
    if token==None:
      return None
    child1 = self._ParseStochasticExpression()
    if child1==None:
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_THEN)
    if token==None:
      return None
    child2 = self._ParseStochasticExpression()
    if child2==None:
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_ELSE)
    if token==None:
      return None
    child3 = self._ParseStochasticExpression()
    if child3==None:
      return None
    return S2MLStochasticExpression.NewOperation(S2MLType.IF_EXPRESSION, [child1, child2, child3])

  def _ParseWithinExpression(self):
    token = self._ParseGivenToken(S2MLToken.TOKEN_LPAR)
    if token==None:
      return None
    token = self._ParseGivenToken(S2MLToken.KEYWORD_WITHIN)
    if token==None:
      return None
    child = self._ParseStochasticExpression()
    if child==None:
      return None
    domain = self._ParseDomain()
    if domain==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    return S2MLStochasticExpression.NewWithin(child, domain)

  def _ParseOperatorCall(self):
    functor = self._ParseIdentifier()
    if functor==None:
      return None
    children = self._ParseStochasticExpressionList()
    if children==None:
      return None
    return S2MLStochasticExpression.NewOperatorCall(functor, children)

  def _ParseStochasticExpressionList(self):
    token = self._ParseGivenToken(S2MLToken.TOKEN_LPAR)
    if token==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLToken.TOKEN_RPAR:
      self._currentTokenIndex += 1
      return []
    expressions = []
    while True:
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      expressions.append(expression)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    return expressions

  def _ParseDistribution(self):
    token = self._ParseGivenToken(S2MLToken.TOKEN_LBRACK)
    if token==None:
      return None
    pairs = []
    while True:
      pair = self._ParsePair()
      if pair==None:
        return None
      pairs.append(pair)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=S2MLToken.TOKEN_COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(S2MLToken.TOKEN_RBRACK)
    if token==None:
      return None
    distribution = S2MLDistribution()
    distribution.AddPairs(pairs)
    return distribution

  def _ParsePair(self):
    token = self._ParseGivenToken(S2MLToken.TOKEN_LPAR)
    if token==None:
      return None
    x = self._ParseNumber()
    if x==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_COMMA)
    if token==None:
      return None
    y = self._ParseNumber()
    if y==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    return (x, y)

# 3.7) Functions and operators
# ----------------------------

  def _ParseFunction(self, container):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_FUNCTION)
    if not self._model.CanComposeFunction(container):
      self._RaiseParseError(token, "Function declarations are not authorized here")
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLToken.KEYWORD_END)
      return None
    functionDeclaration = self._model.NewFunction(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLToken.TOKEN_EQUAL:
        break
      identifier = self._ParseIdentifier()
      if identifier==None:
        return None
      functionDeclaration.AppendArgument(identifier)
    token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
    if token==None:
      return None
    return self._ParseBlockBody(functionDeclaration)

  def _ParseFunctionCall(self, container):
    if not self._model.CanComposeFunctionCall(container):
      self._RaiseParseError(token, "Function calls are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_LPAR)
    if token==None:
      return None
    functionIdentifier = self._ParseIdentifier()
    if functionIdentifier==None:
      return None
    directive = self._model.NewFunctionCall(functionIdentifier, container)
    while True:
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      directive.AppendArgument(expression)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLToken.TOKEN_RPAR:
        break
    token = self._ParseGivenToken(S2MLToken.TOKEN_RPAR)
    if token==None:
      return None
    token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    return directive

  def _ParseOperator(self, container):
    token = self._ParseGivenToken(S2MLToken.KEYWORD_OPERATOR)
    if not self._model.CanComposeOperator(container):
      self._RaiseParseError(token, "Operator declarations are not authorized here")
      self._RecoverError(S2MLToken.TOKEN_SEMICOLON)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLToken.TOKEN_SEMICOLON)
      return None
    operatorDeclaration = self._model.NewOperator(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLToken.TOKEN_EQUAL:
        break
      identifier = self._ParseIdentifier()
      if identifier==None:
        return None
      operatorDeclaration.AppendArgument(identifier)
    token = self._ParseGivenToken(S2MLToken.TOKEN_EQUAL)
    if token==None:
      return None
    definition = self._ParseStochasticExpression()
    if definition==None:
      return None
    operatorDeclaration.SetDefinition(definition)
    token = self._ParseGivenToken(S2MLToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    return operatorDeclaration

# 4) Auxiliary functions
# ----------------------

  def _ParseIdentifier(self):
    token = self._ParseGivenToken(S2MLToken.TOKEN_IDENTIFIER)
    if token==None:
      return None
    return token.GetString()

  def _ParsePath(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None or (token.GetType()!=S2MLToken.TOKEN_IDENTIFIER and token.GetType()!=S2MLToken.TOKEN_PATH):
      self._RaiseParseError(token, "A path was expected")
      return None
    self._currentTokenIndex += 1
    return S2MLPath.FromString(token.GetString())

  def _ParseToken(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token, "Unexpected end of file")
      return None
    self._currentTokenIndex += 1
    return token

  def _ParseGivenToken(self, tokenType):
    token = self._ParseToken()
    if token==None:
      return None
    if token.GetType()!=tokenType:
      self._RaiseParseError(token, "A %s was expected" % self._lexer.GetStringOfCode(tokenType))
      return None
    return token

  def _ParseNumber(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token, "Unexpected end of file")
      return None
    if token.GetType()==S2MLToken.TOKEN_INTEGER:
      self._currentTokenIndex += 1
      return int(token.GetString())
    if token.GetType()==S2MLToken.TOKEN_REAL:
      self._currentTokenIndex += 1
      return float(token.GetString())
    self._RaiseParseError(token, "A number was expected")
    return None

  def _RecoverError(self, tokenType):
    currentTokenIndex = self._currentTokenIndex
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      self._currentTokenIndex += 1
      if token.GetType()==tokenType:
        break
    if currentTokenIndex==self._currentTokenIndex:
      sys.exit()

  def _RaiseParseError(self, token, message):
    self._numberOfErrors += 1
    fullMessage = ""
    if token!=None:
      fullMessage += "Parse error line %d (%d, %s):\n" % (token.GetLineNumber(), token.GetColumnNumber(), token.GetString())
    fullMessage += message
    self._engine.RaiseError(fullMessage)

