# S2MLCore version 1.2.0
# S2MLPath.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creators
# 3) Accessors

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

# 2) Creators
# -----------

class S2MLPath:
  COMPOSITION_SEPARATOR = "."
  PSEUDO_COMPOSITION_SEPARATOR = "_"
  ROOT_IDENTIFIER = "main"
  CONTAINER_IDENTIFIER = "owner"
  SELF_IDENTIFIER = "self"

  def __init__(self):
    self._identifiers = []
    self._string = None

  @staticmethod
  def NewEmptyPath():
    return S2MLPath()
    
  @staticmethod
  def FromString(string):
    if string=="":
      return S2MLPath.NewEmptyPath()
    identifiers = string.split(S2MLPath.COMPOSITION_SEPARATOR)
    for identifier in identifiers:
      if not S2MLPath.IsValidIdentifier(identifier):
        return None
    self = S2MLPath()
    self._identifiers = identifiers
    self._string = string
    return self

  @staticmethod
  def Copy(path):
    self = S2MLPath()
    for identifier in path._identifiers:
      self.AppendIdentifier(identifier)
    if path._string!=None:
      self._string = path._string
    return self

# 3) Accessors
# ------------

  @staticmethod
  def Merge(path1, path2):
    self = S2MLPath()
    self._identifiers = path1._identifiers + path2._identifiers
    return self

  @staticmethod
  def IsValidIdentifier(string):
    return re.match("^[a-zA-Z_][a-zA-Z0-9_-]*$", string)

  @staticmethod
  def IsValidPath(string):
    identifiers = string.split(S2MLPath.COMPOSITION_SEPARATOR)
    for identifier in identifiers:
      if not S2MLPath.IsValidIdentifier(identifier):
        return False
    return True

  def ToString(self):
    if self._string!=None:
      return self._string
    self._string = ""
    for identifier in self._identifiers:
      if self._string=="":
        self._string = identifier
      else:
        self._string += S2MLPath.COMPOSITION_SEPARATOR + identifier
    return self._string

  def ToIdentifier(self):
    string = ""
    for identifier in self._identifiers:
      if string=="":
        string = identifier
      else:
        string += S2MLPath.PSEUDO_COMPOSITION_SEPARATOR + identifier
    return string

  def MakeIdentifier(self):
    identifier = self.ToString()
    self._string = identifier
    self._identifiers = [identifier]

  def IsEmpty(self):
    return self.GetSize()==0

  def GetSize(self):
    return len(self._identifiers)

  def IsIdentifier(self):
    return self.GetSize()==1

  def IsLocal(self):
    for identifier in self._identifiers:
      if identifier==S2MLPath.ROOT_IDENTIFIER or identifier==S2MLPath.CONTAINER_IDENTIFIER:
        return False
    return True

  def IsAbsolute(self):
    for identifier in self._identifiers:
      if identifier==S2MLPath.ROOT_IDENTIFIER:
        return True
    return False

  def GetLastIdentifier(self):
    return self.GetIdentifier(self.GetSize()-1)

  def GetFirstIdentifier(self):
    return self.GetIdentifier(0)

  def GetIdentifier(self, index):
    if index<0 or index>=len(self._identifiers):
      return None
    return self._identifiers[index]

  def AppendIdentifier(self, identifier):
    self._identifiers.append(identifier)
    self._string = None

  def PrependIdentifier(self, identifier):
    self._identifiers.insert(0, identifier)
    self._string = None

  def RemoveLastIdentifier(self):
    self._string = None
    return self._identifiers.pop()

  def RemoveFirstIdentifier(self):
    self._string = None
    return self._identifiers.pop(0)

  def GetIdentifiers(self):
    return self._identifiers

  def Resolve(self):
    oldIdentifiers = self._identifiers
    self._identifiers = []
    for identifier in oldIdentifiers:
      if identifier==S2MLPath.ROOT_IDENTIFIER:
        self._identifiers = []
        self._identifiers.append(S2MLPath.ROOT_IDENTIFIER)
      elif identifier==S2MLPath.CONTAINER_IDENTIFIER:
        if len(self._identifiers)==0:
          self._identifiers = oldIdentifiers
          return False
        self._identifiers.pop()
      elif identifier==S2MLPath.SELF_IDENTIFIER:
        self._identifiers.pop()
      else:
        self._identifiers.append(identifier)
    return True

  def GetSuffix(self, other):
    if len(self._identifiers)<len(other._identifiers):
      return None
    for i in range(0, len(other._identifiers)):
      if self.GetIdentifier(i)!=other.GetIdentifier(i):
        return None
    result = S2MLPath()
    for i in range(len(other._identifiers), len(self._identifiers)):
      result.AppendIdentifier(self.GetIdentifier(i))
    return result

