# S2MLCore version 1.2.0
# S2MLEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Accessors
# 4) Scripts
# 5) Builtins
# 6) Options

# 1) Imported modules
# -------------------

import pathlib

from S2MLModels import *
from S2MLInterface import *
from S2MLStatistics import *

from S2MLModelChecker import *
from S2MLModelFlattener import *
from S2MLModelRewriter import *
from S2MLModelManager import *

from S2MLScriptPrinter import *
from S2MLScriptParser import *
from S2MLScriptChecker import *
from S2MLInterpreter import *

from S2MLEnvironmentManager import *

from S2MLStochasticExpressionManager import *


# 2) Creator
# ----------

class S2MLEngine:

  # Execution statuses
  COMPLETED = 0
  ACTIVE = 1
  CANCELLED = 2

  def __init__(self):
    self._currentWorkingDirectory = pathlib.Path('.')
    self._modelManager = None
    self._interpreter = None
    self._environmentManager = None
    self._expressionManager = None
    self._forceFlattening = False
    self._numberOfErrors = 0
    self._firstFreeMatrixIndex = 0
    self._matrix = None
    self._executionStatus = S2MLEngine.COMPLETED

# 3) Accessors
# ------------

  def CreateTools(self):
    self._modelManager = S2MLModelManager(self)
    self._interpreter = S2MLInterpreter(self)
    self._environmentManager = S2MLEnvironmentManager(self)
    self._modelManager.CreateTools()
    self._interpreter.CreateTools()
    self._environmentManager.CreateTools()
    self._environmentManager.Initialize()
    self._expressionManager = S2MLStochasticExpressionManager(self)

  def GetCurrentWorkingDirectory(self):
    return self._currentWorkingDirectory

  def SetCurrentWorkingDirectory(self, directoryName):
    self._currentWorkingDirectory = self.GetFilePath(directoryName)

  def GetFilePath(self, fileName):
    filePath = pathlib.Path(fileName)
    if not filePath.is_absolute():
      filePath = self._currentWorkingDirectory / filePath
    filePath = filePath.resolve()
    return filePath

  def GetInterpreter(self):
    return self._interpreter

  def GetScriptPrinter(self):
    return self._interpreter.GetScriptPrinter()

  def GetModelManager(self):
    return self._modelManager

  def GetEnvironmentManager(self):
    return self._environmentManager

  def GetExpressionManager(self):
    return self._expressionManager

  def GetMatrix(self):
    return self._matrix

  def OpenSession(self):
    self.CreateTools()
    return 0

  def CloseSession(self):
    self._interpreter.FlushOutputAndErrorFiles()
    return 0

  def GetNumberOfErrors(self):
    return self._numberOfErrors

  def RaiseError(self, message):
    self._numberOfErrors += 1
    self._interpreter.PrintErrorMessage(message)

  def RaiseErrorInCommand(self, command, message):
    if command.GetFileName()!=None:
      fullMessage = "In file '{0:s}', line {1:d}:\n{2:s}".format(command.GetFileName(), command.GetLineNumber(), message)
    else:
      fullMessage = "At line {0:d}:\n{1:s}".format(command.GetLineNumber(), message)
    self.RaiseError(fullMessage)

  def GetExecutionStatus(self):
    return self._executionStatus

  def SetExecutionStatus(self, status):
    self._executionStatus = status

  def GetOutputFile(self):
    return self._interpreter.GetOutputFile()

# 4) Scripts
# ----------

  def LoadScript(self, fileName):
    self._interpreter.LoadScript(fileName)

  def ExecuteScript(self, script):
    error = 0
    for command in script.GetCommands():
      error = self.ExecuteCommand(command)
      if error!=0:
        break
    return error

  def ExecuteCommand(self, command):
    if command.GetType()==S2MLCommand.SCRIPT:
      error = self.ExecuteScript(command)
    elif command.GetType()==S2MLCommand.BUILTIN:
      error = self.ExecuteBuiltInCommand(command)
    else:
      self.RaiseErrorInCommand("Unknow command\n")
      error = 1
    return error

  def ExecuteBuiltInCommand(self, command):
    commandName = command.GetName()
    if commandName=="compute":
      return self.ExecuteCommandCompute(command)
    if commandName=="instantiate":
      return self.ExecuteCommandInstantiate(command)
    if commandName=="flatten":
      return self.ExecuteCommandFlatten(command)
    if commandName=="load":
      return self.ExecuteCommandLoad(command)
    if commandName=="print":
      return self.ExecuteCommandPrint(command)
    if commandName=="reset":
      return self.ExecuteCommandReset(command)
    if commandName=="set":
      return self.ExecuteCommandSet(command)
    if commandName=="profile":
      return self.ExecuteCommandProfile(command)
    self.RaiseErrorInCommand("Unknow command {1:s}\n".format(commandName))
    return 1

# 5) Builtins
# -----------

# 5.1) compute
# ------------

  def ExecuteCommandCompute(self, command):
    self.RaiseErrorInCommand(command, "Not implemented\n")
    return 1

# 5.2) load
# ---------

  # load model fileName
  # load script fileName
  def ExecuteCommandLoad(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="model":
      fileName = command.GetArgumentAsString(1)
      error = self._modelManager.ImportModel(fileName)
      return error
    if what=="script":
      fileName = command.GetArgumentAsString(1)
      error = self._interpreter.ImportScript(fileName)
      return error
    self.RaiseErrorInCommand(command, "Incorrect argument {0:s}\n".format(what))
    return 1

# 5.3) instantiate and flatten
# ----------------------------

  # instantiate model
  def ExecuteCommandInstantiate(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="model":
      error = self._modelManager.InstantiateModel(False)
      return error
    return 1

  # flatten model
  def ExecuteCommandFlatten(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="model":
      error = self._modelManager.InstantiateModel(True)
      return error
    return 1

# 5.4) print
# ----------

  # print model [output=fileName mode=(append|write)]
  # print instantiated-model [output=fileName mode=(append|write)]
  # print variables [output=fileName mode=(append|write)]
  # print variable variableName [output=fileName mode=(append|write)]
  # print profiles [output=fileName mode=(append|write)]
  # print profile profileName [output=fileName mode=(append|write)]
  def ExecuteCommandPrint(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="model":
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._modelManager.PrintModel(self._interpreter.GetOutputFile())
      else:
        error = self._modelManager.ExportModel(outputFileName, mode)
      return error
    if what=="instantiated-model":
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._modelManager.PrintInstantiatedModel(self._interpreter.GetOutputFile())
      else:
        error = self._modelManager.ExportInstantiatedModel(outputFileName, mode)
      return error
    if what=="variables":
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._environmentManager.PrintEnvironmentVariables(self._interpreter.GetOutputFile())
      else:
        error = self._environmentManager.ExportEnvironmentVariables(outputFileName, mode)
      return error
    if what=="variable":
      variableName = command.GetArgumentAsIdentifier(1)
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._environmentManager.PrintEnvironmentVariable(variableName, self._interpreter.GetOutputFile())
      else:
        error = self._environmentManager.ExportEnvironmentVariable(variableName, outputFileName, mode)
      return error
    if what=="profiles":
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._environmentManager.PrintIndicatorProfiles(self._interpreter.GetOutputFile())
      else:
        error = self._environmentManager.ExportIndicatorProfiles(outputFileName, mode)
      return error
    if what=="profile":
      profileName = command.GetArgumentAsIdentifier(1)
      outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
      mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
      if outputFileName==None:
        error = self._environmentManager.PrintIndicatorProfile(profileName, self._interpreter.GetOutputFile())
      else:
        error = self._environmentManager.ExportIndicatorProfile(profileName, outputFileName, mode)
      return error
    self.RaiseErrorInCommand(command, "Incorrect argument {0:s}\n".format(what))
    return 1

# 5.5) reset
# ----------

  # reset model
  def ExecuteCommandReset(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="model":
      self.ResetModel()
      return 0
    if what=="target-block":
      self._modelManager.ResetTargetBlock()
      return 0
    identifier = what
    error = self._environmentManager.ResetValueOfEnvironmentVariable(identifier)
    return error

# 5.6) set
# --------

  # set target-block blockName
  # set variable value
  def ExecuteCommandSet(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="target-block":
      blockName = command.GetArgumentAsIdentifier(1)
      error = self._modelManager.SetTargetBlock(blockName)
      return error
    identifier = what
    value = command.GetArgument(1)
    error = self._environmentManager.SetValueOfEnvironmentVariable(identifier, value)
    return error

# 5.7) profile
# ------------

  def ExecuteCommandProfile(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="new":
      sourceName = command.GetArgumentAsIdentifier(1)
      error = self._environmentManager.NewIndicatorProfile(sourceName)
      return error
    if what=="clone":
      sourceName = command.GetArgumentAsIdentifier(1)
      targetName = command.GetArgumentAsIdentifier(2)
      error = self._environmentManager.CloneIndicatorProfile(sourceName, targetName)
      return error
    if what=="delete":
      sourceName = command.GetArgumentAsIdentifier(1)
      error = self._environmentManager.DeleteIndicatorProfile(sourceName)
      return error
    if what=="set":
      sourceName = command.GetArgumentAsIdentifier(1)
      argument = command.GetArgumentAsIdentifier(2)
      if argument=="mean":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionMean(sourceName, value)
        return error
      if argument=="standard-deviation":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionStandardDeviation(sourceName, value)
        return error
      if argument=="confidence-interval-90":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionConfidenceInterval90(sourceName, value)
        return error
      if argument=="confidence-interval-95":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionConfidenceInterval95(sourceName, value)
        return error
      if argument=="confidence-interval-99":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionConfidenceInterval99(sourceName, value)
        return error
      if argument=="extrema":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionExtrema(sourceName, value)
        return error
      if argument=="quantiles":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionQuantiles(sourceName, value)
        return error
      if argument=="number-of-quantiles":
        value = command.GetArgumentAsInteger(3)
        error = self._environmentManager.SetIndicatorProfileOptionNumberOfQuantiles(sourceName, value)
        return error
      if argument=="distribution":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionDistribution(sourceName, value)
        return error
      if argument=="cumulative-distribution":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionCumulativeDistribution(sourceName, value)
        return error
      if argument=="number-of-points":
        value = command.GetArgumentAsInteger(3)
        error = self._environmentManager.SetIndicatorProfileOptionNumberOfPoints(sourceName, value)
        return error
      if argument=="bins":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionBins(sourceName, value)
        return error
      if argument=="number-of-bins":
        value = command.GetArgumentAsInteger(3)
        error = self._environmentManager.SetIndicatorProfileOptionNumberOfBins(sourceName, value)
        return error
      if argument=="bootstrapping":
        value = command.GetArgumentAsBoolean(3)
        error = self._environmentManager.SetIndicatorProfileOptionBootstrapping(sourceName, value)
        return error
      if argument=="sampling-ratio":
        value = command.GetArgumentAsReal(3)
        error = self._environmentManager.SetIndicatorProfileOptionSamplingRatio(sourceName, value)
        return error
      if argument=="number-of-samples":
        value = command.GetArgumentAsInteger(3)
        error = self._environmentManager.SetIndicatorProfileOptionNumberOfSamples(sourceName, value)
        return error
      self.RaiseErrorInCommand(command, "Incorrect argument {0:s}\n".format(argument))
      return 1
    self.RaiseErrorInCommand(command, "Incorrect argument {0:s}\n".format(what))
    return 1

# 6) Options
# ----------

  def GetValueOfOptionOfCommand(self, command, name):
    value = command.GetOption(name)
    if value==None:
      value = self._environmentManager.GetValueOfEnvironmentVariable(name)
    return value

  def GetValueOfOptionOfCommandAsBoolean(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsBoolean():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be a Boolean\n".format(name))
      return None
    return value.GetValueOfBoolean()

  def GetValueOfOptionOfCommandAsInteger(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsInteger():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be an integer\n".format(name))
      return None
    return value.GetValueOfInteger()

  def GetValueOfOptionOfCommandAsReal(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsNumber():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be a floating point number\n".format(name))
      return None
    return value.GetValueOfReal()

  def GetValueOfOptionOfCommandAsIdentifier(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsIdentifier():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be an identifier\n".format(name))
      return None
    return value.GetValueOfIdentifier()

  def GetValueOfOptionOfCommandAsString(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsString():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be a string\n".format(name))
      return None
    return value.GetValueOfString()

  def GetValueOfOptionOfCommandAsListOfReals(self, command, name, defaultValue = None):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return defaultValue
    if not value.IsListOfReals():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be a list of floating point numbers\n".format(name))
      return None
    return value.GetValueOfListOfReals()

  def GetValueOfOptionOfCommandAsFileOpeningMode(self, command, name):
    value = self.GetValueOfOptionOfCommand(command, name)
    if value==None:
      return None
    if not value.IsIdentifier():
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be an opening mode (write or append)\n".format(name))
      return None
    mode = value.GetValueOfIdentifier()
    if mode=="write":
      mode = "w"
    if mode=="append":
      mode = "a"
    if mode!="w" and mode!="a":
      self.RaiseErrorInCommand(command, "The value of option '{0:s}' must be an opening mode (write or append)\n".format(name))
      return None
    return mode

  def GetValueOfOptionOfCommandAsTargetBlock(self, command):
    targetBlockName = self.GetValueOfOptionOfCommandAsIdentifier(command, "target-block")
    if targetBlockName==None:
      targetBlock = self._modelManager.SelectForTargetBlock()
    else:
      targetBlock = self._modelManager.LookForTargetBlock(targetBlockName)
    return targetBlock