# S2MLCore version 1.2.0
# S2MLModelChecker.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Undefined references
# 4) Circular definitions
# 5) Types
# 6) Tool specific checks

# 1) Imported modules
# -------------------

from S2MLModels import *

# 2) Creator
# ----------

class S2MLModelChecker:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._model = None
    self._checkedElement = None
    self._checkedBlock = None
    self._currentBlock = None
    self._stack = []
    self._count = 0
    self._errors = 0

  def RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

  def CheckModel(self, model):
    self.CheckReferences(model)
    if self._errors!=0:
      return self._errors
    self.CheckCircularDefinitions(model)
    if self._errors!=0:
      return self._errors
    self.ToolSpecificCheckModel(model)
    return self._errors

# 3) Undefined references
# -----------------------

  def CheckReferences(self, model):
    self._model = model
    self._errors = 0
    self.CheckReferencesInBlock(self._model)
    return self._errors

  def CheckReferencesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==S2MLType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==S2MLType.PARAMETER:
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
    return self._errors

  def _CheckReferencesInAttributesOfElement(self, element):
    for name in element.GetAttributeNames():
      value = element.GetAttribute(name)
      self._CheckReferencesInStochasticExpression(value)

  def _CheckReferencesInStochasticExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referencePath = expression.GetReferencePath()
      element = self._checkedBlock.AccessNamedElement(referencePath)
      if element==None:
        referenceIdentifier = referencePath.ToString()
        element = self._checkedBlock.LookForIdentifiedElement(referenceIdentifier)
      if element==None:
        checkedElementPath = self._checkedElement.GetPath()
        self.RaiseError("In block %s, undefined reference %s in definition of %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString(), checkedElementPath.ToString()))
        return
      if not self._model.IsValidReferredElementInStochasticExpression(element):
        checkedElementPath = self._checkedElement.GetPath()
        if checkedElementPath==None:
          self.RaiseError("In block %s, incorrect reference %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString()))
        else:
          self.RaiseError("In block %s, incorrect reference %s in definition of %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString(), checkedElementPath.ToString()))
        return
      expression.SetReferredElement(element)
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      pass
    else:
      for child in expression.GetChildren():
        self._CheckReferencesInStochasticExpression(child)

# 4) Circular definitions
# -----------------------

  def CheckCircularDefinitions(self, model):
    self._model = model
    self._errors = 0
    self.CheckCircularDefinitionsInBlock(self._model)
    self._model.ResetFlagsOfElements()
    return self._errors

  def CheckCircularDefinitionsInBlock(self, block):
    previousCheckedBlock = self._checkedBlock
    self._checkedBlock = block
    previousBlock = self._currentBlock
    self._currentBlock = block
    for element in block.GetElements():
      if element.GetFlag()==0:
        if element.GetType()==S2MLType.BLOCK:
          self.CheckCircularDefinitionsInBlock(element)
        elif element.GetType()==S2MLType.PARAMETER:
          self.CheckCircularDefinitionsInDefinition(element)
    self._checkedBlock = previousCheckedBlock
    self._currentBlock = previousBlock
    return self._errors

  def CheckCircularDefinitionsInDefinition(self, element):
    if element.GetFlag()==2:
      return
    if element.GetFlag()==1:
      message = "In block %s, circular definition of %s\n" % (self._checkedBlock.GetAbsolutePath().ToString(), element.GetPath().ToString())
      for involvedParameter in self._stack:
        message += involvedParameter.GetIdentifier() + "\n"
      self.RaiseError(message)
      return
    definition = element.GetDefinition()
    if definition==None:
      self.RaiseError("In block %s, undefined element %s\n" % (self._checkedBlock.GetAbsolutePath().ToString(), element.GetPath().ToString()))
      return
    element.SetFlag(1)
    self._stack.append(element)
    self.CheckCircularDefinitionsInStochasticExpression(definition)
    self._stack.pop()
    element.SetFlag(2)

  def CheckCircularDefinitionsInStochasticExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referencePath = expression.GetReferencePath()
      element = self._currentBlock.AccessNamedElement(referencePath)
      if element==None:
        element = self._currentBlock.LookForIdentifiedElement(referencePath.ToString())
      if element==None:
        self.RaiseError("Internal error %s %s" % (self._currentBlock.GetAbsolutePath().ToString(), referencePath.ToString()))
        return
      if element.GetType()==S2MLType.PARAMETER:
        previousBlock = self._currentBlock
        self._currentBlock = element.GetContainer()
        self.CheckCircularDefinitionsInDefinition(element)
        self._currentBlock = previousBlock
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      pass
    else:
      for child in expression.GetChildren():
        self.CheckCircularDefinitionsInStochasticExpression(child)

# 5) Types
# --------

  def CheckTypes(self, model):
    self._model = model
    self._errors = 0
    self.CheckTypesInBlock(self._model)
    return self._errors

  def CheckTypesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==S2MLType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      else:
        self._checkedElement = element
        self.CheckTypesInElement(element)
        self._checkedElement = None
    return self._errors

  def CheckTypesInElement(self, element):
    if element.GetType()==S2MLType.PARAMETER:
      baseType = self.CheckTypesInStochasticExpression(element.GetDefinition())

  def CheckTypesInParameterDeclaration(self, parameter):
    baseType = self.CheckTypesInStochasticExpression(parameter.GetDefinition())
    self.CheckBaseTypeAgainstDomain(baseType, parameter.GetDomain())

  def CheckBaseTypeAgainstDomain(self, baseType, domain):
    pass

  def CheckTypesInStochasticExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referencePath = expression.GetReferencePath()
      element = self._currentBlock.AccessNamedElement(referencePath)
      if element==None:
        element = self._currentBlock.LookForIdentifiedElement(referencePath.ToString())
      if element==None:
        baseType = S2MLConstant.NONE
      else:
        baseType = self.GetBaseTypeOfElement(element)
    elif expression.GetType()==S2MLType.CONSTANT:
      baseType = expression.GetValue().GetType()
    elif expression.GetType()==S2MLType.UNIFORM_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.NORMAL_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.LOGNORMAL_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.TRIANGULAR_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.EXPONENTIAL_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.WEIBULL_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.PIECEWISE_UNIFORM_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.STEPWISE_DEVIATE_EXPRESSION \
      or expression.GetType()==S2MLType.CONSTANT_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.EXPONENTIAL_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.GLM_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.WEIBULL_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.DIRAC_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.PIECEWISE_UNIFORM_DISTRIBUTION_EXPRESSION \
      or expression.GetType()==S2MLType.STEPWISE_DISTRIBUTION_EXPRESSION:
      baseType = S2MLConstant.REAL
    elif expression.GetType()==S2MLType.ADD_EXPRESSION \
      or expression.GetType()==S2MLType.SUB_EXPRESSION \
      or expression.GetType()==S2MLType.MUL_EXPRESSION \
      or expression.GetType()==S2MLType.DIV_EXPRESSION \
      or expression.GetType()==S2MLType.NEG_EXPRESSION \
      or expression.GetType()==S2MLType.MAX_EXPRESSION \
      or expression.GetType()==S2MLType.LT_EXPRESSION \
      or expression.GetType()==S2MLType.GT_EXPRESSION \
      or expression.GetType()==S2MLType.LEQ_EXPRESSION \
      or expression.GetType()==S2MLType.GEQ_EXPRESSION:
      baseType = S2MLConstant.INTEGER
      for child in expression.Children():
        childBaseType = self.CheckTypesInStochasticExpression(child)
        if childBaseType==S2MLConstant.INTEGER:
          pass
        elif childBaseType==S2MLConstant.REAL:
          baseType = S2MLConstant.REAL
        else:
          self.RaiseError("Type error in relation with %s\n" % self._checkedElement.GetAbsolutePath().ToString())
          self._errors += 1
      if expression.GetType()==S2MLType.LT_EXPRESSION \
        or expression.GetType()==S2MLType.GT_EXPRESSION \
        or expression.GetType()==S2MLType.LEQ_EXPRESSION \
        or expression.GetType()==S2MLType.GEQ_EXPRESSION:
        baseType = S2MLConstant.BOOLEAN
    elif expression.GetType()==S2MLType.AND_EXPRESSION \
      or expression.GetType()==S2MLType.OR_EXPRESSION \
      or expression.GetType()==S2MLType.NOT_EXPRESSION \
      or expression.GetType()==S2MLType.COUNT_EXPRESSION:
      baseType = S2MLConstant.BOOLEAN
      for child in expression.Children():
        childBaseType = self.CheckTypesInStochasticExpression(child)
        if childBaseType!=S2MLConstant.BOOLEAN:
          self.RaiseError("Type error in relation with %s\n" % self._checkedElement.GetAbsolutePath().ToString())
          self._errors += 1
      if expression.GetType()==S2MLType.COUNT_EXPRESSION:
        baseType = S2MLConstant.INTEGER
    elif expression.GetType()==S2MLType.EQ_EXPRESSION or expression.GetType()==S2MLType.DF_EXPRESSION:
      leftChild = expression.GetChild(0)
      rightChild = expression.GetChild(1)
      leftBaseType = self.CheckTypesInStochasticExpression(leftChild)
      rightChildType = self.CheckTypesInStochasticExpression(rightChild)
      if leftBaseType==S2MLConstant.INTEGER:
        leftBaseType = S2MLConstant.REAL
      if rightChildType==S2MLConstant.INTEGER:
        rightChildType = S2MLConstant.REAL
      if leftBaseType!=rightBaseType:
        self.RaiseError("Type error in relation with %s\n" % self._checkedElement.GetAbsolutePath().ToString())
        self._errors += 1
      baseType = S2MLConstant.BOOLEAN
    elif expression.GetType()==S2MLType.WITHIN_EXPRESSION:
      baseType = S2MLConstant.BOOLEAN
    elif expression.GetType()==S2MLType.IF_EXPRESSION:
      baseType = S2MLConstant.ANY
      ifChild = expression.GetChild(0)
      thenChild = expression.GetChild(1)
      elseChild = expression.GetChild(2)
      ifBaseType = self.CheckTypesInStochasticExpression(ifChild)
      if ifBaseType!=S2MLConstant.BOOLEAN:
        self.RaiseError("Type error in relation with %s\n" % self._checkedElement.GetAbsolutePath().ToString())
        self._errors += 1
        baseType = S2MLConstant.NONE
      thenChildType = self.CheckTypesInStochasticExpression(thenChild)
      elseChildType = self.CheckTypesInStochasticExpression(thenChild)
      if leftBaseType!=rightBaseType:
        self.RaiseError("Type error in relation with %s\n" % self._checkedElement.GetAbsolutePath().ToString())
        self._errors += 1
        if baseType==S2MLConstant.ANY:
          baseType = S2MLConstant.NONE
      else:
        if baseType==S2MLConstant.ANY:
          baseType = thenChildType
    else:
      baseType = S2MLConstant.NONE
    return baseType

  def GetBaseTypeOfElement(self, element):
    if element.GetType()==S2MLType.PARAMETER:
      return element.GetBaseType()
    return S2MLConstant.ANY

# 6) Tool specific checks
# -----------------------

  def ToolSpecificCheckModel(self, model):
    pass
