# S2MLCore version 1.2.0
# S2MLScriptPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Scripts and Commands
# 4) Environment variables
# 5) Indicator profiles
# 6) Managed values

# 1) Imported modules
# -------------------

from S2MLManagedValue import *
from S2MLCommand import *

# 2) Creator
# ----------

class S2MLScriptPrinter:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._errors = 0

  def _RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

# 3) Methods to export scripts
# ----------------------------

  def ExportScript(self, script, fileName):
    try:
      file = open(fileName, "w")
    except:
      self._RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintScript(script, file)
    file.close()

  def PrintScript(self, script, file):
    for command in script.GetCommands():
      self.PrintCommand(command, file)

  def PrintCommand(self, command, file):
    if command.GetType()==S2MLCommand.SCRIPT:
      self.PrintScript(command, file)
    elif command.GetType()==S2MLCommand.BUILTIN:
      self.PrintBuiltInCommand(command, file)

  def PrintBuiltInCommand(self, command, file):
    file.write(command.GetName())
    for argument in command.GetArguments():
      file.write(" ")
      self.PrintManagedValue(argument, file)
    for name in command.GetOptions():
      file.write(" ")
      value = command.GetOption(name)
      file.write(name)
      file.write("=")
      self.PrintManagedValue(value, file)
    file.write("\n")

# 4) Environment variables
# ------------------------

  def ExportEnvironmentVariables(self, environmentVariables, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintEnvironmentVariables(environmentVariables, output)
    output.close()
    return 0

  def PrintEnvironmentVariables(self, environmentVariables, output):
    for identifier in environmentVariables.keys():
      self.PrintEnvironmentVariable(identifier, environmentVariables[identifier], output)


  def ExportEnvironmentVariable(self, variableName, value, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintEnvironmentVariable(variableName, value, output)
    output.close()
    return 0

  def PrintEnvironmentVariable(self, identifier, value, output):
    output.write("set ")
    output.write(identifier)
    output.write(" ")
    self.PrintConstant(value, output)
    output.write("\n")

# 5) Indicator profiles
# ---------------------

  def ExportIndicatorProfiles(self, indicateProfiles, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintIndicatorProfiles(indicateProfiles, output)
    output.close()
    return 0

  def PrintIndicatorProfiles(self, indicateProfiles, output):
    for indicateProfile in indicateProfiles:
      self.PrintIndicatorProfile(indicateProfile, output)


  def ExportIndicatorProfile(self, indicateProfile, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintIndicatorProfile(indicateProfile, output)
    output.close()
    return 0

  def PrintIndicatorProfile(self, indicateProfile, output):
    self._PrintIndicatorProfileBooleanOption("mean", indicateProfile.GetName(), indicateProfile.GetMean(), output)
    self._PrintIndicatorProfileBooleanOption("standard-deviation", indicateProfile.GetName(), indicateProfile.GetStandardDeviation(), output)
    self._PrintIndicatorProfileBooleanOption("confidence-interval-90", indicateProfile.GetName(), indicateProfile.GetConfidenceRange90(), output)
    self._PrintIndicatorProfileBooleanOption("confidence-interval-95", indicateProfile.GetName(), indicateProfile.GetConfidenceRange95(), output)
    self._PrintIndicatorProfileBooleanOption("confidence-interval-99", indicateProfile.GetName(), indicateProfile.GetConfidenceRange99(), output)
    self._PrintIndicatorProfileBooleanOption("extrema", indicateProfile.GetName(), indicateProfile.GetExtrema(), output)
    self._PrintIndicatorProfileBooleanOption("quantiles", indicateProfile.GetName(), indicateProfile.GetQuantiles(), output)
    self._PrintIndicatorProfileIntegerOption("number-of-quantiles", indicateProfile.GetName(), indicateProfile.GetNumberOfQuantiles(), output)
    self._PrintIndicatorProfileBooleanOption("distribution", indicateProfile.GetName(), indicateProfile.GetDistribution(), output)
    self._PrintIndicatorProfileBooleanOption("cumulative-distribution", indicateProfile.GetName(), indicateProfile.GetCumulativeDistribution(), output)
    self._PrintIndicatorProfileIntegerOption("number-of-points", indicateProfile.GetName(), indicateProfile.GetNumberOfPoints(), output)
    self._PrintIndicatorProfileBooleanOption("bins", indicateProfile.GetName(), indicateProfile.GetBins(), output)
    self._PrintIndicatorProfileIntegerOption("number-of-bins", indicateProfile.GetName(), indicateProfile.GetNumberOfBins(), output)
    self._PrintIndicatorProfileIntegerOption("number-of-stored-values", indicateProfile.GetName(), indicateProfile.GetNumberOfStoredValues(), output)
    self._PrintIndicatorProfileBooleanOption("bootstrapping", indicateProfile.GetName(), indicateProfile.GetBootstrapping(), output)
    self._PrintIndicatorProfileIntegerOption("number-of-samples", indicateProfile.GetName(), indicateProfile.GetNumberOfSamples(), output)
    self._PrintIndicatorProfileFloatOption("sampling-ratio", indicateProfile.GetName(), indicateProfile.GetSamplingRatio(), output)

  def _PrintIndicatorProfileBooleanOption(self, optionName, indicatorName, optionValue, output):
    output.write("profile set ")
    output.write(optionName)
    output.write(" ")
    output.write(indicatorName)
    output.write(" ")
    self.PrintBoolean(optionValue, output)
    output.write("\n")

  def _PrintIndicatorProfileIntegerOption(self, optionName, indicatorName, optionValue, output):
    output.write("profile set ")
    output.write(optionName)
    output.write(" ")
    output.write(indicatorName)
    output.write(" ")
    self.PrintInteger(optionValue, output)
    output.write("\n")

  def _PrintIndicatorProfileFloatOption(self, optionName, indicatorName, optionValue, output):
    output.write("profile set ")
    output.write(optionName)
    output.write(" ")
    output.write(indicatorName)
    output.write(" ")
    self.PrintFloat(optionValue, output)
    output.write("\n")


# 4) Managed values
# -----------------

  def PrintManagedValue(self, value, file):
    if value.GetType()==S2MLManagedValue.NONE:
      pass
    elif value.GetType()==S2MLManagedValue.BOOLEAN:
      if value.GetValueOfBoolean():
        file.write("true")
      else:
        file.write("false")
    elif value.GetType()==S2MLManagedValue.INTEGER:
      file.write("{0:d}".format(value.GetValueOfInteger()))
    elif value.GetType()==S2MLManagedValue.REAL:
      file.write("{0:g}".format(value.GetValueOfReal()))
    elif value.GetType()==S2MLManagedValue.SYMBOL:
      file.write("{0:s}".format(value.GetValueOfSymbol()))
    elif value.GetType()==S2MLManagedValue.STRING:
      file.write("{0:s}".format(value.GetValueOfString()))
    elif value.GetType()==S2MLManagedValue.LIST:
      file.write("[")
      first = True
      for childValue in value.GetValueOfList():
        if first:
          first = False
        else:
          file.write(", ")
        self.PrintManagedValue(childValue, file)
      file.write("]")
